
#include "common.hxx"
#include "basic.hxx"
#include "Config.hxx"
#include "Address.hxx"
#include "Input.hxx"
#include "Output.hxx"
#include "LongUDPSocket.hxx"
#include "ObjectPool.hxx"
#include "objects.hxx"
using namespace Rescue;

// by koto
int fread2(S32* buffer, int size, int count, FILE* fp) {
	ASSERT(size == sizeof(S32) && count == 1);
	
	Byte b1;
	if(fread(&b1, sizeof(Byte), 1, fp) != 1)
		return 0;
	Byte b2;
	if(fread(&b2, sizeof(Byte), 1, fp) != 1)
		return 0;
	Byte b3;
	if(fread(&b3, sizeof(Byte), 1, fp) != 1)
		return 0;
	Byte b4;
	if(fread(&b4, sizeof(Byte), 1, fp) != 1)
		return 0;
	
	*buffer = (S32)b1 | ((S32)b2 << 8) | ((S32)b3 << 16) | ((S32)b4 << 24);
	return 1;
}
//

int main2(int argc, char** argv);
void makeWorld(ObjectPool& pool, int number);

int main(int argc, char** argv)
{
	try {
		return main2(argc, argv);
	} catch(std::exception& e) {
		fprintf(stderr, "%s\n", (const char*)e.what());
	} catch(...) {
		fprintf(stderr, "error...\n");
	}
	return 1;
}

int main2(int argc, char** argv)
{
	const char* host = "localhost";
	Config config;
	int number = 50;

	// 
	int i;
	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			if(strcmp(argv[i], "-file") == 0) {
				if(++i < argc) {
					config.readConfigFile(argv[i]);
				} else {
					fprintf(stderr, "error: '-file'\n");
				}
			} else if(strcmp(argv[i], "-n") == 0) {
				if(++i < argc) {
					number = atoi(argv[i]);
				} else {
					fprintf(stderr, "error: '-file'\n");
				}
			} else {
				if(++i < argc) {
					if(!config.setValue(argv[i-1]+1, argv[i]))
						fprintf(stderr, "error: '%s'\n", argv[i]);
				} else {
					fprintf(stderr, "error: '%s'\n", argv[i-1]);
				}
			}
		} else {
			fprintf(stderr, "error: '%s'\n", argv[i]);
		}
	}

	// \Pbg쐬
	Address address = Address(INADDR_ANY, config.gis_port(), true);
	LongUDPSocket socket(address, config.textlogname_gis());
	Input input;
	Output output;
	output.setWait(config.send_udp_wait());
	Address to(host, config.port());	//M

	printf("start\n");

	// bZ[W[v
	S32 time = 0;
	ObjectPool pool;
	makeWorld(pool, number);
	for(;;) {
		try {
			socket.receive(input);	// TODO: G[OŕԂ	//f[^MAinput ֒~B
			for(;;) {
				Header header = input.get();	//input (4oCg)f[^oAheader ֑
				if(header == HEADER_NULL)
					break;

				S32 size = input.get();
				Input::Cursor start = input.cursor();	//ݓǂݏoĂʒuL

				switch(header) {
				default:
					printf("Unknown header. %lx\n", (long)header);
					break;
				case KG_CONNECT:
					printf("KG_CONNECT\n");
					to = socket.addressRecievedFrom();
					output.clear();
					output.put(GK_CONNECT_OK);
					output.put(~(S32)0);
					pool.output(output, 0);
					output.put(HEADER_NULL);
					socket.send(output, to);
					break;
				case KG_ACKNOWLEDGE:
					printf("KG_ACKNOWLEDGE\n");
					break;
				case KG_UPDATE:
					printf("KG_UPDATE\n");
					time = input.get();
					pool.restructure(time, input);
					break;
				}

				input.setCursor(start);
				input.skip(size);						//size oCgXLbv
			}
		} catch(std::exception& e) {
			fprintf(stderr, "%s\n", (const char*)e.what());
		} catch(...) {
			fprintf(stderr, "error...\n");
		}
	}
	
	return 0;
}

// GIS by M.Hatayama
void makeWorld(ObjectPool& pool, int number)
{
	FILE *		fpRoad;
	FILE *		fpNode;
	FILE *		fpBuilding;
	int			result;
	S32			nCntRoad, nCntNode, nCntBuilding;
	S32			nCntRec, nCntRec2;
	S32			tmp, loop1, loop2, tmp_byte;
	World *		pWorld;
	Road *		pRoad;
	Node *		pNode;
	Building *	pBuilding;
	Bytes		bytes;
	
	bool		bFire = false;

	//world data
	pWorld = new World;
	pool.testAdd( pWorld );

	//load road data
	if( (fpRoad = fopen("road.bin", "r+b") ) == NULL){
		printf("cannot find file --- road.bin\n");
		return;
	}
	result = fseek( fpRoad, 0L, SEEK_SET);
	result = fseek( fpRoad, sizeof( S32 )*3, SEEK_CUR);
	result = fread2( &nCntRoad, sizeof( S32 ), 1, fpRoad );
	pRoad = new Road[nCntRoad];
printf("Road:%ld\n", nCntRoad );

	//load noad data
	if( (fpNode = fopen("node.bin", "r+b") ) == NULL){
		printf("cannot find file --- node.bin\n");
		fclose(fpRoad);
		return;
	}
	result = fseek( fpNode, 0L, SEEK_SET);
	result = fseek( fpNode, sizeof( S32 )*3, SEEK_CUR);
	result = fread2( &nCntNode, sizeof( S32 ), 1, fpNode );
	pNode = new Node[nCntNode];
printf("Node:%ld\n", nCntNode );

	//load building data
	if( (fpBuilding = fopen("building.bin", "r+b") ) == NULL){
		printf("cannot find file --- road.bin\n");
		fclose(fpRoad);
		fclose(fpNode);
		return;
	}
	result = fseek( fpBuilding, 0L, SEEK_SET);
	result = fseek( fpBuilding, sizeof( S32 )*3, SEEK_CUR);
	result = fread2( &nCntBuilding, sizeof( S32 ), 1, fpBuilding );
	pBuilding = new Building[nCntBuilding];
printf("Building:%ld\n", nCntBuilding );
	for(int i=0 ;i<4; i++) {
		Refuge* refuge = (Refuge*)&pBuilding[nCntBuilding/4*i];
		new(refuge) Refuge();	// BAD!BAD!BAD!BAD!
	}

	//make road data
	for ( loop1=0; loop1<nCntRoad; loop1++ ){
		result = fread2( &nCntRec, sizeof( S32 ), 1, fpRoad );
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setId( tmp );
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pNode[tmp-1].testConnect(0, &pRoad[loop1], true);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pNode[tmp-1].testConnect(0, &pRoad[loop1], false);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setLength(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setRoadKind(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setCarsPassToHead(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setCarsPassToTail(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setHumansPassToHead(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setHumansPassToTail(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setWidth(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setBlock(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setRepairCost(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setMedianStrip(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setLinesToHead(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setLinesToTail(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpRoad );
		pRoad[loop1].setWidthForWalkers(0, tmp);

		pool.testAdd( &pRoad[loop1] );
	}
	fclose(fpRoad);

	//make node data
	for ( loop1=0; loop1<nCntNode; loop1++ ){
		result = fread2( &nCntRec, sizeof( S32 ), 1, fpNode );
		result = fread2( &tmp, sizeof( S32 ), 1, fpNode );
		pNode[loop1].setId( tmp );
		result = fread2( &tmp, sizeof( S32 ), 1, fpNode );
		pNode[loop1].setX(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpNode );
		pNode[loop1].setY(0, tmp);

		result = fread2( &nCntRec2, sizeof( S32 ), 1, fpNode );
		for (loop2=0; loop2<nCntRec2; loop2++){
			result = fread2( &tmp, sizeof( S32 ), 1, fpNode );
			//	eIDs & bIDs
		}

		result = fread2( &tmp, sizeof( S32 ), 1, fpNode );
		pNode[loop1].setSignal(0, tmp);

		bytes.clear();
		for (loop2=0; loop2<nCntRec2; loop2++){
			result = fread2( &tmp, sizeof( S32 ), 1, fpNode );
			tmp_byte = tmp & 0xff000000;
			bytes.push_back( (Byte)(tmp_byte >> 24) );
			tmp_byte = tmp & 0x00ff0000;
			bytes.push_back( (Byte)(tmp_byte >> 16) );
			tmp_byte = tmp & 0x0000ff00;
			bytes.push_back( (Byte)(tmp_byte >> 8) );
			tmp_byte = tmp & 0x000000ff;
			bytes.push_back( (Byte)tmp_byte );
		}
		pNode[loop1].setShortcutToTurn(0, bytes);

		bytes.clear();
		for (loop2=0; loop2<nCntRec2; loop2++){
			result = fread2( &tmp, sizeof( S32 ), 1, fpNode );
			tmp_byte = tmp & 0xff000000;
			bytes.push_back( (Byte)(tmp_byte >> 24) );
			tmp_byte = tmp & 0x00ff0000;
			bytes.push_back( (Byte)(tmp_byte >> 16) );
			tmp_byte = tmp & 0x0000ff00;
			bytes.push_back( (Byte)(tmp_byte >> 8) );
			tmp_byte = tmp & 0x000000ff;
			bytes.push_back( (Byte)tmp_byte );

			result = fread2( &tmp, sizeof( S32 ), 1, fpNode );
			tmp_byte = tmp & 0xff000000;
			bytes.push_back( (Byte)(tmp_byte >> 24) );
			tmp_byte = tmp & 0x00ff0000;
			bytes.push_back( (Byte)(tmp_byte >> 16) );
			tmp_byte = tmp & 0x0000ff00;
			bytes.push_back( (Byte)(tmp_byte >> 8) );
			tmp_byte = tmp & 0x000000ff;
			bytes.push_back( (Byte)tmp_byte );
		}
		pNode[loop1].setPocketToTurnAcross(0, bytes);

		bytes.clear();
		for (loop2=0; loop2<nCntRec2; loop2++){
			result = fread2( &tmp, sizeof( S32 ), 1, fpNode );
			tmp_byte = tmp & 0xff000000;
			bytes.push_back( (Byte)(tmp_byte >> 24) );
			tmp_byte = tmp & 0x00ff0000;
			bytes.push_back( (Byte)(tmp_byte >> 16) );
			tmp_byte = tmp & 0x0000ff00;
			bytes.push_back( (Byte)(tmp_byte >> 8) );
			tmp_byte = tmp & 0x000000ff;
			bytes.push_back( (Byte)tmp_byte );

			result = fread2( &tmp, sizeof( S32 ), 1, fpNode );
			tmp_byte = tmp & 0xff000000;
			bytes.push_back( (Byte)(tmp_byte >> 24) );
			tmp_byte = tmp & 0x00ff0000;
			bytes.push_back( (Byte)(tmp_byte >> 16) );
			tmp_byte = tmp & 0x0000ff00;
			bytes.push_back( (Byte)(tmp_byte >> 8) );
			tmp_byte = tmp & 0x000000ff;
			bytes.push_back( (Byte)tmp_byte );

			result = fread2( &tmp, sizeof( S32 ), 1, fpNode );
			tmp_byte = tmp & 0xff000000;
			bytes.push_back( (Byte)(tmp_byte >> 24) );
			tmp_byte = tmp & 0x00ff0000;
			bytes.push_back( (Byte)(tmp_byte >> 16) );
			tmp_byte = tmp & 0x0000ff00;
			bytes.push_back( (Byte)(tmp_byte >> 8) );
			tmp_byte = tmp & 0x000000ff;
			bytes.push_back( (Byte)tmp_byte );
		}
		pNode[loop1].setSignalTiming(0, bytes);

//printf("b:%ld, %ld,%ld\n", pNode[loop1].id(), pNode[loop1].x() ,pNode[loop1].y() );		
		pool.testAdd( &pNode[loop1] );
	}
	fclose(fpNode);

	//make building data
	for ( loop1=0; loop1<nCntBuilding; loop1++ ){
//printf("%ld/%ld\n", loop1, nCntBuilding);
		result = fread2( &nCntRec, sizeof( S32 ), 1, fpBuilding );
		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
		pBuilding[loop1].setId( tmp );
		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
		pBuilding[loop1].setX(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
		pBuilding[loop1].setY(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
		pBuilding[loop1].setFloors(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
		pBuilding[loop1].setBuildingAttributes(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
//		if (bFire != TRUE){
//			tmp = 1;
//			printf("%ld, %ld\n", pBuilding[loop1].x(), pBuilding[loop1].y() );
//			bFire = TRUE;
//		}
		if (pBuilding[loop1].x() == 23046400 && pBuilding[loop1].y() == 3731000 ){
			tmp = 1;
		}
		if (pBuilding[loop1].x() == 22776000 && pBuilding[loop1].y() == 3621000 ){
			tmp = 1;
		}
		if (pBuilding[loop1].x() == 22942100 && pBuilding[loop1].y() == 3796300 ){
			tmp = 1;
		}
		if (pBuilding[loop1].x() == 22975800 && pBuilding[loop1].y() == 3702000 ){
			tmp = 1;
		}
		pBuilding[loop1].setIgnition(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
		pBuilding[loop1].setFieryness(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
		pBuilding[loop1].setBrokenness(0, tmp);

		result = fread2( &nCntRec2, sizeof( S32 ), 1, fpBuilding );
		for (loop2=0; loop2<nCntRec2; loop2++){
			result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
			pNode[tmp-1].testConnect(0, &pBuilding[loop1]);
		}

		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
//		pBuilding[loop1].setBuildingShapeId(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
		pBuilding[loop1].setBuildingAreaGround(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
		pBuilding[loop1].setBuildingAreaTotal(0, tmp);
		result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
		pBuilding[loop1].setBuildingCode(0, tmp);

		result = fread2( &nCntRec2, sizeof( S32 ), 1, fpBuilding );
		bytes.clear();
		for (loop2=0; loop2<nCntRec2; loop2++){

			result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
			tmp_byte = tmp & 0xff000000;
			bytes.push_back( (Byte)(tmp_byte >> 24) );
			tmp_byte = tmp & 0x00ff0000;
			bytes.push_back( (Byte)(tmp_byte >> 16) );
			tmp_byte = tmp & 0x0000ff00;
			bytes.push_back( (Byte)(tmp_byte >> 8) );
			tmp_byte = tmp & 0x000000ff;
			bytes.push_back( (Byte)tmp_byte );

			result = fread2( &tmp, sizeof( S32 ), 1, fpBuilding );
			tmp_byte = tmp & 0xff000000;
			bytes.push_back( (Byte)(tmp_byte >> 24) );
			tmp_byte = tmp & 0x00ff0000;
			bytes.push_back( (Byte)(tmp_byte >> 16) );
			tmp_byte = tmp & 0x0000ff00;
			bytes.push_back( (Byte)(tmp_byte >> 8) );
			tmp_byte = tmp & 0x000000ff;
			bytes.push_back( (Byte)tmp_byte );
		}
		pBuilding[loop1].setBuildingApexes(0, bytes);
		
		pool.testAdd( &pBuilding[loop1] );
	}
	fclose(fpBuilding);
	
	int k, z;

//	k = number;
//	z = nCntRoad / k;
//	if(z < 1)
//		z = 1;
//	for ( loop1=0; loop1<nCntRoad; loop1++ ){
//		if(loop1 % z == 0 && k-- > 0) {
//			Civilian* civilian = new Civilian();
//			pool.testAdd(civilian);
//			civilian->setPosition(0, &pRoad[loop1]);
//			civilian->setPositionExtra(0, pRoad[loop1].length() * 3 / 10);
//		}
//	}

	// Civilian in Building
	S32 CivilianNum;
	S32 CivilianRatio = 20;

	CivilianNum = nCntBuilding/CivilianRatio;
	for ( loop1=0; loop1<CivilianNum; loop1++ ){
		Civilian* civilian = new Civilian();
		pool.testAdd(civilian);
		civilian->setPosition(0, &pBuilding[loop1*CivilianRatio]);
	}
printf("Civilian in Building:%ld\n", CivilianNum );

	// Civilian on Road
	if ( nCntRoad/CivilianRatio == 0 )
		CivilianNum = 1;
	else
		CivilianNum = nCntBuilding/CivilianRatio;

	for ( loop1=0; loop1<CivilianNum; loop1++ ){
		Civilian* civilian = new Civilian();
		pool.testAdd(civilian);
		civilian->setPosition(0, &pRoad[loop1*CivilianRatio]);
		civilian->setPositionExtra(0, pRoad[loop1*CivilianRatio].length() * 3 / 10);
	}
printf("Civilian on Road:%ld\n", CivilianNum );

	//AmbulanceTeam on Node
	S32 AmbulanceTeamNum;
	if ( nCntNode/50 == 0 )
		AmbulanceTeamNum = 1;
	else if ( nCntNode/50 > 5 )
		AmbulanceTeamNum = 5;
	else
		AmbulanceTeamNum = nCntNode/50;

	for ( loop1=0; loop1<AmbulanceTeamNum; loop1++ ){
		AmbulanceTeam* ambulance = new AmbulanceTeam();
		pool.testAdd(ambulance);
		ambulance->setPosition(0, &pNode[loop1]);
	}
printf("AmbulanceTeam on Node:%ld\n", AmbulanceTeamNum );

	//FireBrigade on Node
	S32 FireBrigadeNum;
	if ( nCntNode/50 == 0 )
		FireBrigadeNum = 1;
	else if ( nCntNode/50 > 10 )
		FireBrigadeNum = 10;
	else
		FireBrigadeNum = nCntNode/50;

	for ( loop1=0; loop1<FireBrigadeNum; loop1++ ){
		FireBrigade* fireteam = new FireBrigade();
		pool.testAdd(fireteam);
		fireteam->setPosition(0, &pNode[nCntNode-loop1-1]);
	}
printf("FireBrigade on Node:%ld\n", FireBrigadeNum );

	//PoliceForce on Node
	S32 PoliceForceNum;
	if ( nCntNode/50 == 0 )
		PoliceForceNum = 1;
	else if ( nCntNode/50 > 10 )
		PoliceForceNum = 10;
	else
		PoliceForceNum = nCntNode/50;

	for ( loop1=0; loop1<PoliceForceNum; loop1++ ){
		PoliceForce* policeforce = new PoliceForce();
		pool.testAdd(policeforce);
		policeforce->setPosition(0, &pRoad[nCntRoad-loop1-1]);
		policeforce->setPositionExtra(0, pRoad[nCntRoad-loop1-1].length() * 6 / 10);
	}
printf("PoliceForce on Node:%ld\n", PoliceForceNum );

//#if 0
//	k = number;
//	z = nCntNode / k;
//	if(z < 1)
//		z = 1;
//	for ( loop1=0; loop1<nCntNode; loop1++ ){
//		if(loop1 % z == 0 && k-- > 0) {
//			AmbulanceTeam* ambulance = new AmbulanceTeam();
//			pool.testAdd(ambulance);
//			ambulance->setPosition(0, &pNode[loop1]);
//		}
//	}
//#endif
}

