// RescueSystem.cxx
//
/////////////////////////////////////////////////////////////////////////////

#include "common.hxx"
#include "RescueSystem.hxx"

#ifndef NDEBUG
#  include "RescueSystem.inl"
#endif

#include "objects.hxx"

namespace Rescue
{
	/////////////////////////////////////////////////////////////////////////
	// RescueSystem
	
	RescueSystem::~RescueSystem()
	{
	}

	void RescueSystem::loop()
	{
		m_objectPool.clearFiredObjects();
		System::loop();
	}
	void RescueSystem::sendToAngents()
	{
		//long theTime = ::timeGetTime();

		// TODO: 
		// TODO: M鏇Ԃ̓_]܂B
		const double squaredVision = (double)config().vision() * config().vision();
		for(unsigned int i=0; i < m_agentTypes.size(); i++) {
#if 1	// b
			receiveMessages(agentSocket());
#endif
			Agents::iterator it = m_pConnectedAgents[i].begin();
			for(; it != m_pConnectedAgents[i].end(); it++) {
				m_outputBuffer.clear();
				m_outputBuffer.put(KA_SENSE);
				Output::Cursor base = m_outputBuffer.put(~(S32)0);
				Agent* agent = *it;
				ASSERT(dynamic_cast<RealObject*>(agent->asObject()) != 0);
				RealObject* self = (RealObject*)agent->asObject();
				m_outputBuffer.put(self->id());
				m_outputBuffer.put(m_time);
				// self
				self->outputObject(m_outputBuffer);
				// map (苗)
				RescueObjectPool::Enumerator e = m_objectPool.objectsInRound(self->x(), self->y(), config().vision());
				Object* o;
				while(o = e.next(), o != 0)
					o->outputObject(m_outputBuffer);
				// mapi΍Ёj
				Objects::const_iterator it2 = m_objectPool.firedObjects().begin();
				for(; it2 != m_objectPool.firedObjects().end(); it2++) {
					Building* b = dynamic_cast<Building*>(*it2);
					if(b != 0) {
						double x = (double)b->x() - self->x();
						double y = (double)b->y() - self->y();
						if(x * x + y * y > squaredVision)
							b->outputObject(m_outputBuffer);
					}
				}
				m_outputBuffer.put(TYPE_NULL);
				m_outputBuffer.setSize(base);
				m_outputBuffer.put(HEADER_NULL);
				agentSocket().send(m_outputBuffer, agent->clientAddress(), m_time);
			}
		}

		//printf("sendToAgent: %ld\n", ::timeGetTime() - theTime);

		m_objectPool.clearFiredObjects();
	}
	
	void RescueSystem::processCommand(Agent* agent, Header header, S32 size, Input& input, const LongUDPSocket& from)
	{
		Humanoid* h = dynamic_cast<Humanoid*>(agent);
		if(h == 0 || h->hp() > 0) {	// {ȂǐlłȂAlŎłȂH
			if(header != AK_SAY && header != AK_TELL) {
				bool block = false;	// R}hubN邩H
				switch(header) {
				case AK_MOVE:
					if(h == 0 || h->buriedness() > 0)	// lłȂA܂ĂH
						block = true;
					break;
				case AK_EXTINGUISH:
					{
						// TODO: 삩ꍇoĂȂB
						FireBrigade* f = dynamic_cast<FireBrigade*>(agent);
						if(f == 0) {
							block = true;
						} else {
#if 0	// ^N͍lȂ
							S32 value = f->waterQuantity();
							Input::Cursor cursor = m_inputBuffer.cursor();
							S32 target;
							while(target = m_inputBuffer.get(), target != 0) {
								m_inputBuffer.skip(4*3);
								S32 x = m_inputBuffer.get();
								if(x < 0)
									block = true;
								value -= x;
							}
							if(value < 0)	//TODO: 镪͕B
								block = true;
							else
								f->setWaterQuantity(m_time, value);
							m_inputBuffer.setCursor(cursor);
#endif
						}
					}
					// TODO: ȂǂłtB^O
					break;
				}
				if(!block) {
					System::processCommand(agent, header, size, input, from);
				}
			} else {
				// TODO: ׂ̂ȂƂB
				// TODO: ̂̃^[̍ŏɂB
				// TODO: sȂ炵ׂ邪Aǂ悤B
				// TODO: lɂȂǂ邩B
				S32 size = m_inputBuffer.get();
				if(size <= config().say_max_bytes()) {
					S32 sizeBound = (size + 3) & ~3;
					Byte* buffer = new Byte[sizeBound+1];
					m_inputBuffer.get(sizeBound, buffer);
					buffer[size] = '\0';
					S32 target = m_inputBuffer.get();
					//printf("say '%s'\n", buffer);
					Id sender = agent->asObject()->id();
					say(sender, target, (char*)buffer);	// TODO: LXgĂ邪AƂB
					delete[] buffer;
				}
			}
		}
	}

	void RescueSystem::say(Id from, Id target, char* message)
	{
		Agent* agent = dynamic_cast<Agent*>(m_objectPool.get(target));
		if(agent) {
			m_outputBuffer.clear();
			m_outputBuffer.put(KA_HEAR);
			Output::Cursor base = m_outputBuffer.put(~(S32)0);
			m_outputBuffer.put(target);
			m_outputBuffer.put(from);
			m_outputBuffer.putString(message);
			m_outputBuffer.setSize(base);
			m_outputBuffer.put(HEADER_NULL);
			//printf("say %s\n", message);
			agentSocket().send(m_outputBuffer, agent->clientAddress(), m_time);
		}
	}
	
	/////////////////////////////////////////////////////////////////////////
} // namespace Rescue
